// tinyasm.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "main.h"
#include "assembler.h"

#pragma warning(disable: 4996)

char* sourceFilename = NULL;
char* outputFilename = NULL;
char* outputType = NULL;

int main(int argc, char* argv[])
{
	if (!ParseCommandLine(argc, argv))
	{
		return -1;
	}

	FILE *pSource = fopen(sourceFilename, "r");
	if (!pSource)
	{
		printf("Could not open source file %s\n", sourceFilename);
		return -1;
	}

    FILE *pOutput;
    pOutput = fopen(outputFilename, "wb");

	if (!pOutput)
	{
		printf("Could not open output file %s\n", outputFilename);
		return -1;
	}

	Assembler assembler;
	uint8vec outputBuffer;

	if (assembler.Assemble(sourceFilename, pSource, &outputBuffer))
	{
		// write the output file
		int outputSize = outputBuffer.size();

		for (int i=0; i<outputSize; i++)
		{
			OutputBin(pOutput, outputBuffer[i]);
		}
		for (int i=0; i<Assembler::kAddressSpace - outputSize; i++)
		{
			OutputBin(pOutput, 0);
		}

		printf("binary output: wrote %s (%d program bytes, %d%% full)\n", outputFilename, outputSize, 100 * outputSize / Assembler::kAddressSpace);      

		FILE *pSymbols;
		string symbolFilename = outputFilename;
		int idx = symbolFilename.find_last_of('.');
		if (idx > 0)
		{
			symbolFilename = symbolFilename.substr(0, idx) + ".sym";
		}
		else
		{
			symbolFilename += ".sym";
		}

		pSymbols = fopen(symbolFilename.c_str(), "w");

		if (!pSymbols)
		{
			printf("Could not open output symbol file %s\n", symbolFilename);
			return -1;
		}

		for (Assembler::LabelToAddressMap::iterator it = assembler.m_labels.begin(); it != assembler.m_labels.end(); ++it)
		{
			string lineToWrite = it->first;
			lineToWrite += "  =$";
			char hexNumber[10];
			sprintf(hexNumber, "%X ;\n", it->second);
			lineToWrite += hexNumber;
			fputs(lineToWrite.c_str(), pSymbols);
		}

		fclose(pSymbols);
		printf("symbols: wrote %s\n", symbolFilename.c_str());
	}

    fclose(pSource);
	fclose(pOutput);

	return 0;
}

void OutputBin(FILE* pFile, int val)
{
    fputc(val, pFile);
}

bool ParseCommandLine(int argc, char* argv[])
{
	for (int i=1; i < argc; i++)
	{
		char* arg = argv[i];

		if (arg[0] == '-')
		{
			switch(arg[1])
			{
			case 'o':
				if (i < argc-1)
				{
					outputFilename = argv[i+1];
					i++;
				}
				break;

			default:
				printf("Unrecognized option %s\n", arg);
				Usage(argv[0]);
				return false;
			}
		}
		else
		{
			if (sourceFilename == NULL)
			{
				sourceFilename = arg;
			}
			else
			{
				printf("Unrecognized option %s\n", arg);
				Usage(argv[0]);
				return false;
			}
		}
	}

	if (sourceFilename == NULL)
	{
		printf("No source file specified.\n");
		Usage(argv[0]);
		return false;
	}

	if (outputFilename == NULL)
	{
		printf("No output file specified.\n");
		Usage(argv[0]);
		return false;
	}

	return true;
}

void Usage(char* name)
{
	printf("%s -o outfile infile\n", name);
}
