﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace MicrocodeBuilder
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private void runButton_Click(object sender, EventArgs eventArgs)
        {
            Dictionary<int, int> microcode = new Dictionary<int, int>();

            // initialize everything to zero
            for (int i = 0; i < 2048; i++)
                microcode[i] = 0;

            // inputs
            // 0 0 0 0 0 0 instr3 instr2 instr1 instr0 /C /Z phase

            // outputs
            // hi: incPC /loadPC     /loadA /loadFlags | /carryIn     M     S3         S2 
            // lo:    S1      S0     /csRAM     /weRAM |   /oeALU /oeIn /oeOprnd /loadOut

            // phase 0 is the same for all instructions: fetch
            for (int instr = 0; instr < 16; instr++)
            {
                for (int cz = 0; cz < 4; cz++)
                {
                    microcode[(instr << 3) | (cz << 1) | 0] = 0xF837;
                }
            }

            // 0 JC
            for (int z = 0; z < 2; z++)
            {
                microcode[(0 << 3) | (0 << 2) | (z << 1) | 1] = 0x3837; // C
                microcode[(0 << 3) | (1 << 2) | (z << 1) | 1] = 0xF837; // NC
            }

            // 1 JNC
            for (int z = 0; z < 2; z++)
            {
                microcode[(1 << 3) | (0 << 2) | (z << 1) | 1] = 0xF837; // C
                microcode[(1 << 3) | (1 << 2) | (z << 1) | 1] = 0x3837; // NC
            }

            // 2 CMPI
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(2 << 3) | (cz << 1) | 1] = 0x61BD;
            }

            // 3 CMPM
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(3 << 3) | (cz << 1) | 1] = 0xE19F;
            }

            // 4 LIT
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(4 << 3) | (cz << 1) | 1] = 0x46BD;
            }

            // 5 IN
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(5 << 3) | (cz << 1) | 1] = 0x46BB;
            }

            // 6 LD
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(6 << 3) | (cz << 1) | 1] = 0xC69F;
            }

            // 7 ST
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[((7 << 3) | cz << 1) | 1] = 0xF807;
            }

            // 8 JZ
            for (int c = 0; c < 2; c++)
            {
                microcode[(8 << 3) | (c << 2) | (0 << 1) | 1] = 0x3837; // Z
                microcode[(8 << 3) | (c << 2) | (1 << 1) | 1] = 0xF837; // NZ
            }

            // 9 JNZ
            for (int c = 0; c < 2; c++)
            {
                microcode[(9 << 3) | (c << 2) | (0 << 1) | 1] = 0xF837; // Z
                microcode[(9 << 3) | (c << 2) | (1 << 1) | 1] = 0x3837; // NZ
            }

            // 10 ADDI
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(10 << 3) | (cz << 1) | 1] = 0x4A7D; 
            }

            // 11 ADDM
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(11 << 3) | (cz << 1) | 1] = 0xCA5F; 
            }

            // 12 JMP
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(12 << 3) | (cz << 1) | 1] = 0x3037;
            }

            // 13 OUT
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(13 << 3) | (cz << 1) | 1] = 0x7836;
            }

            // 14 NORI
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(14 << 3) | (cz << 1) | 1] = 0x447D;
            }

            // 15 NORM
            for (int cz = 0; cz < 4; cz++)
            {
                microcode[(15 << 3) | (cz << 1) | 1] = 0xC45F;
            }



            try
            {
                FileStream fs_low = File.Create("microcode_0.bin", 2048, FileOptions.None);
                BinaryWriter bw_low = new BinaryWriter(fs_low);

                FileStream fs_high = File.Create("microcode_1.bin", 2048, FileOptions.None);
                BinaryWriter bw_high = new BinaryWriter(fs_high);

                for (int i = 0; i < 2048; i++)
                {
                    bw_low.Write((byte)(microcode[i] & 0xFF));
                    bw_high.Write((byte)((microcode[i] >> 8) & 0xFF));
                }

                bw_low.Close();
                fs_low.Close();
                bw_high.Close();
                fs_high.Close();

                MessageBox.Show("Wrote microcode_0.bin and microcode_1.bin");
                Application.Exit();
            }
            catch (Exception e)
            {
                Console.Write(e.Message);
                Console.ReadKey(true);
            }
        }
    }
}
