﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace Music_Maker
{
    public partial class Form1 : Form
    {
        int tempo = 120;
        int octave = 4;
        int octaveOffset = 0;
        int defaultNoteLength = 8;
        int noteLength = 8;
        float noteLengthScale = 1.0f;
        int note = -1;
        float legato = 0.875f;
        int returnAddr = 0;

        public Form1()
        {
            InitializeComponent();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            //A-G [#,+,-]	
            //A-G are notes. # or + following a note produces a sharp; - produces a flat.

            //Any note followed by #,+,or - must refer to a black key on a piano.

            //L(n)	
            //Sets the length of each note. L4 is a quarter note, L1 is a whole note, and so on. n may be from 1 to 64.

            //Length may also follow the note to change the length for that note only. A16 is equivalent to L16A.

            //MF	Music foreground. PLAY and SOUND statements are to run in foreground. That is, each subsequent note or sound is not started until the previous note or sound is finished. This is the initial default.
            //MB	Music background. PLAY and SOUND statements are to run in background. That is, each note or sound is placed in a buffer allowing the BASIC program to continue execution while music plays in the background. As many as 32 notes (or rests) can be played in background at one time.
            //MN	Music normal. Each note plays seven-eighths of the time determined by L (length).
            //ML	Music legato. Each note plays the full period set by L.
            //MS	Music staccato. Each note plays three-quarters of the time determined by L.
            //N(n)	Play note n. n may range from 0 to 84. In the 7 possible octaves, there are 84 notes. n set to 0 indicates a rest.
            //O(n)	Octave 0 sets the current octave. There are 7 octaves (0 through 6). Default is 4. Middle C is at the beginning of octave 3.
            //P(n)	Pause. P may range from 1-64.
            //T(n)	Tempo. T sets the number of L4s in a minute. n may range from 32-255. Default is 120.
            //. (period)	
            //A period after a note increases the playing time of the note by 3/2 times the period determined by L (length of note) times T (tempo). Multiple periods can appear after  a note, and the playing time is scaled accordingly. For example, A. will cause the note A to play one and half times the playing time determined by L (length of the note) times T (the tempo); two periods placed after A (A..) will cause the note to be played at 9/4 times its ascribed value; an A with three periods (A...) at 27/8, etc.

            //Periods may also appear after a P (pause), and increase the pause length as described above.

            //Xstring;	
            //Executes a substring, where string is a variable assigned to a string of PLAY commands.

            //Because of the slow clock interrupt rate, some notes do not play at higher tempos; for example, 1.64 at T255. These note/tempo combinations must be determined through experimentation.

            //>n	A greater-than symbol preceding the note n plays the note in the next higher octave.
            //<n	A less-than symbol preceding the note n plays the note in the next lower octave.

            //string input = "t200 o2 l4 ge2gb2 o3c+ o3d2 e8d8 o2b2." + "p2 b o3 e2ef8e.d o2b8o3c."; // Rising Sun Blues
            
            //string input = "t160 o3" +
            //    "l8 g4eee4ddg4efe4.d" +
            //    "gag4a4gab2 p4ga b4b4b4.b g4b4d4.g " +
            //    "o2b4 o3 d4 e4f4g4.ge4d4 o2b4o3d4e2."; // Little Mathy Groves

            // this one requires some hand editing of the generated file to create repeated phrases
            string input = "DF#A L2 O4 D L4 O5 DD P4" +
                "T180 DF#A L2 A L4 O4 AA P4 F#F# P4 O3 D" +
                "DF#A L2 A L4 O4 AA P4 GG P4 O3 C#" +
                "C#EB L2 B L4 O4 BB P4 GG P4 O3 C#" +
                "C#EB L2 B L4 O4 BB P4 F+F+ P4 O3 D" +
                //110 PLAY PARTTUNE$
                "O4 AA P4 O3 D" +
                //130 PLAY PARTTUNE$
                "O4 BB P4 EEG L8 B P8 ML B1 L4 MN G#A ML L3 O5 F#1" +
                "L4 MN D O4 F# ML L2 F# MN L4 E ML L2 B MN L4 A" +
                "D P8 D8 D4";

            try
            {
                FileStream output = File.Create("music.asm");
                StreamWriter sw = new StreamWriter(output);

                sw.WriteLine("; music.asm");

                FileStream inputStub = File.Open("music-stub.asm", FileMode.Open);
                StreamReader sr = new StreamReader(inputStub);

                while (!sr.EndOfStream)
                {
                    string line = sr.ReadLine();
                    sw.WriteLine(line);
                }

                sr.Close();
                inputStub.Close();

                sw.WriteLine("; the code below was auto geneated by Nibbler Music Maker");
                sw.WriteLine();
                sw.WriteLine("init:");

                while (input.Length > 0)
                {
                    Char cmd = Char.ToUpper(input[0]);
                    input = input.Substring(1);

                    if (cmd == 'N')
                    {
                        EmitNote(sw); 
                        
                        int numberLength = 0;
                        while (Char.IsNumber(input[numberLength]))
                            numberLength++;

                        note = int.Parse(input.Substring(0, numberLength));
                        input = input.Substring(numberLength);
                    }

                    else if (cmd >= 'A' && cmd <= 'G')
                    {
                        EmitNote(sw);

                        if (cmd == 'C')
                            note = octave * 12 + octaveOffset;
                        else if (cmd == 'D')
                            note = octave * 12 + 2 + octaveOffset;
                        else if (cmd == 'E')
                            note = octave * 12 + 4 + octaveOffset;
                        else if (cmd == 'F')
                            note = octave * 12 + 5 + octaveOffset;
                        else if (cmd == 'G')
                            note = octave * 12 + 7 + octaveOffset;
                        else if (cmd == 'A')
                            note = octave * 12 + 9 + octaveOffset;
                        else if (cmd == 'B')
                            note = octave * 12 + 11 + octaveOffset;
                    }

                    else if (cmd == 'P' || cmd == 'R')
                    {
                        EmitNote(sw); 
                        note = 0;
                    }

                    else if (cmd == '#')
                    {
                        if (note != -1)
                            note++;
                        else
                            MessageBox.Show("Unknown command " + cmd);
                    }

                    else if (cmd == '+')
                    {
                        if (note != -1)
                            note++;
                        else
                        {
                            EmitNote(sw);
                            octaveOffset = 12;
                        }
                    }

                    else if (cmd == '>')
                    {
                        if (note != -1)
                            MessageBox.Show("Unknown command " + cmd);
                        else
                        {
                            EmitNote(sw);
                            octaveOffset = 12;
                        }
                    }

                    else if (cmd == '-')
                    {
                        if (note != -1)
                            note--;
                        else
                        {
                            EmitNote(sw);
                            octaveOffset = -12;
                        }
                    }

                    else if (cmd == '<')
                    {
                        if (note != -1)
                            MessageBox.Show("Unknown command " + cmd);
                        else
                        {
                            EmitNote(sw);
                            octaveOffset = -12;
                        }
                    }

                    else if (cmd == '.')
                        noteLengthScale = 1.5f;

                    else if (cmd >= '0' && cmd <= '9')
                    {
                        if (note != -1)
                        {
                            int numberLength = 0;
                            while (input.Length > 0 && Char.IsNumber(input[numberLength]))
                                numberLength++;

                            noteLength = int.Parse(cmd + input.Substring(0, numberLength));
                            input = input.Substring(numberLength);
                        }
                        else
                            MessageBox.Show("Unknown command " + cmd);
                    }

                    else if (cmd == ' ')
                    {
                        EmitNote(sw);
                        continue;
                    }

                    else if (cmd == 'T')
                    {
                        EmitNote(sw); 
                        
                        int numberLength = 0;
                        while (Char.IsNumber(input[numberLength]))
                            numberLength++;

                        tempo = int.Parse(input.Substring(0, numberLength));
                        input = input.Substring(numberLength);                      
                    }

                    else if (cmd == 'L')
                    {
                        EmitNote(sw); 
                        
                        int numberLength = 0;
                        while (Char.IsNumber(input[numberLength]))
                            numberLength++;

                        defaultNoteLength = int.Parse(input.Substring(0, numberLength));
                        noteLength = defaultNoteLength;
                        input = input.Substring(numberLength);
                    }

                    else if (cmd == 'O')
                    {
                        EmitNote(sw); 
                        
                        int numberLength = 0;
                        while (Char.IsNumber(input[numberLength]))
                            numberLength++;

                        octave = int.Parse(input.Substring(0, numberLength));
                        input = input.Substring(numberLength);                       
                    }

                    else if (cmd == 'M')
                    {
                        EmitNote(sw); 
                        
                        input = input.Substring(1);
                        if (Char.ToUpper(input[0]) == 'L')
                            legato = 1.0f; // legato
                        else if (Char.ToUpper(input[0]) == 'S')
                            legato = 0.75f; // staccato
                        else
                            legato = 0.875f; // normal  
                    }

                    else
                    {
                        MessageBox.Show("Unknown command " + cmd);
                    }
                }

                EmitNote(sw);

                // end of music
                sw.WriteLine("jmp halt");

                sw.WriteLine();
                sw.WriteLine("rest_delay_return:");
                for (int i = 0; i < returnAddr; i++)
                {
                    if (i % 16 == 0)
                    {
                        if (i == 0)
                            sw.WriteLine("ld RETURN_ADDR_1");
                        else
                            sw.Write("+ ");
                        if (i != 0)
                            sw.WriteLine(String.Format("cmpi #{0}", i / 16));
                        sw.WriteLine("jnz +");
                        sw.WriteLine("ld RETURN_ADDR_0");
                    }
                    if (i % 16 == 15)
                    {
                        sw.WriteLine("jmp next{0}", i);
                    }
                    else
                    {
                        if (i % 16 != 0)
                            sw.WriteLine(String.Format("cmpi #{0}", i % 16));
                        sw.WriteLine(String.Format("jz next{0}", i));
                    }
                }

                sw.WriteLine("; unknown return address");
                if (returnAddr > 16)
                    sw.WriteLine("+ ");
                sw.WriteLine("halt:");
                sw.WriteLine("jmp halt");

                sw.Close();
                output.Close();

                MessageBox.Show("Wrote music.asm");
                Application.Exit();
            }
            catch (Exception ex)
            {
                Console.Write(ex.Message);
                Console.ReadKey(true);
            }

        }

        void EmitNote(StreamWriter sw)
        {
            if (note != -1)
            {
                // A4 = note 57 = 440 Hz
                double freq = 440 * Math.Pow(2, (note - 57) / 12.0f);
                double periodUs = 1000000.0f / freq + 0.5f;
                int periodUsInt = (int)(periodUs + 0.5f);
                float noteDuration = ((legato * noteLengthScale * 1000000 * 4 * 60.0f / tempo) / noteLength);

                float breakScale = (note == 0) ? 1.0f : 1.0f - legato;
                float breakDuration = ((breakScale * noteLengthScale * 1000000 * 4 * 60.0f / tempo) / noteLength);

                // actual period in microseconds = (48 * (TONE_PERIOD + 1) + 60) / CPU_CLOCKS_PER_US
                // TONE_PERIOD = (desired_period_us * CPU_CLOCKS_PER_US - 108) / 48
                int tonePeriod = (int)((periodUs * 2.4576 - 108) / 48 + 0.5f);
                    
                float noteDurationPeriods = (float)(noteDuration / periodUs);
                int noteDurationPeriodsInt = (int)(noteDurationPeriods + 0.5f);

                float breakDurationDelays = breakDuration * 2.4576f / 100;
                int breakDurationDelaysInt = (int)(breakDurationDelays + 0.5f);

                string noteLetter;
                if (note % 12 == 0)
                    noteLetter = "C";
                else if (note % 12 == 1)
                    noteLetter = "C#";
                else if (note % 12 == 2)
                    noteLetter = "D";
                else if (note % 12 == 3)
                    noteLetter = "D#";
                else if (note % 12 == 4)
                    noteLetter = "E";
                else if (note % 12 == 5)
                    noteLetter = "F";
                else if (note % 12 == 6)
                    noteLetter = "F#";
                else if (note % 12 == 7)
                    noteLetter = "G";
                else if (note % 12 == 8)
                    noteLetter = "G#";
                else if (note % 12 == 9)
                    noteLetter = "A";
                else if (note % 12 == 10)
                    noteLetter = "A#";
                else 
                    noteLetter = "B";

                string noteName = String.Format("{0}{1}", note / 12, noteLetter);

                if (note == 0)
                    sw.WriteLine(String.Format("; Rest duration {0} us ({1} delays = ${2:X})", breakDuration, breakDurationDelaysInt, breakDurationDelaysInt));
                else
                    sw.WriteLine(String.Format("; Play note {0} {10}, freq = {1:f2} Hz, period {4} us, TONE_PERIOD ${5:X}, duration {2} us ({6} periods = ${7:X}), break {3} us ({8} delays = ${9:X})", note, freq, noteDuration, breakDuration, periodUsInt, tonePeriod, noteDurationPeriodsInt, noteDurationPeriodsInt, breakDurationDelaysInt, breakDurationDelaysInt, noteName));

                if (note != 0)
                {
                    sw.WriteLine(String.Format("lit #${0:X}", (tonePeriod >> 8) & 0xF));
                    sw.WriteLine("st TONE_PERIOD_2");
                    sw.WriteLine(String.Format("lit #${0:X}", (tonePeriod >> 4) & 0xF));
                    sw.WriteLine("st TONE_PERIOD_1");
                    sw.WriteLine(String.Format("lit #${0:X}", (tonePeriod >> 0) & 0xF));
                    sw.WriteLine("st TONE_PERIOD_0");

                    sw.WriteLine(String.Format("lit #${0:X}", (noteDurationPeriodsInt >> 8) & 0xF));
                    sw.WriteLine("st TONE_DURATION_2");
                    sw.WriteLine(String.Format("lit #${0:X}", (noteDurationPeriodsInt >> 4) & 0xF));
                    sw.WriteLine("st TONE_DURATION_1");
                    sw.WriteLine(String.Format("lit #${0:X}", (noteDurationPeriodsInt >> 0) & 0xF));
                    sw.WriteLine("st TONE_DURATION_0");
                }

                sw.WriteLine(String.Format("lit #${0:X}", (breakDurationDelaysInt >> 12) & 0xF));
                sw.WriteLine("st DELAY_COUNTER_3");
                sw.WriteLine(String.Format("lit #${0:X}", (breakDurationDelaysInt >> 8) & 0xF));
                sw.WriteLine("st DELAY_COUNTER_2");
                sw.WriteLine(String.Format("lit #${0:X}", (breakDurationDelaysInt >> 4) & 0xF));
                sw.WriteLine("st DELAY_COUNTER_1");
                sw.WriteLine(String.Format("lit #${0:X}", (breakDurationDelaysInt >> 0) & 0xF));
                sw.WriteLine("st DELAY_COUNTER_0");

                sw.WriteLine(String.Format("lit #${0:X}", (returnAddr >> 4) & 0xF));
                sw.WriteLine("st RETURN_ADDR_1");
                sw.WriteLine(String.Format("lit #${0:X}", (returnAddr >> 0) & 0xF));
                sw.WriteLine("st RETURN_ADDR_0");

                if (note != 0)
                    sw.WriteLine("jmp audio_play_tone");
                else
                    sw.WriteLine("jmp rest_delay");

                sw.WriteLine();
                sw.WriteLine(String.Format("next{0}:", returnAddr));

                returnAddr++;
                noteLength = defaultNoteLength; 
                noteLengthScale = 1.0f;
                octaveOffset = 0; 
                note = -1;
            }
        }
    }
}
