; begin music-stub.asm

; OUT ports
#define PORT_CONTROL $E ; 1110 - bit 0 is low

; memory locations
#define TONE_PERIOD_0 $000
#define TONE_PERIOD_1 $001
#define TONE_PERIOD_2 $002
#define TONE_COUNTER_0 $003
#define TONE_COUNTER_1 $004
#define TONE_COUNTER_2 $005
#define TONE_DURATION_0 $006
#define TONE_DURATION_1 $007
#define TONE_DURATION_2 $008
#define DELAY_COUNTER_0 $009
#define DELAY_COUNTER_1 $00A
#define DELAY_COUNTER_2 $00B
#define DELAY_COUNTER_3 $00C
#define RETURN_ADDR_0 $00D
#define RETURN_ADDR_1 $00E

jmp init

rest_delay:
    ; total loop is 50 instructions, 100 clock cycles
    ; first loop is 34 instructions
    lit #11
-   addi #0
    addi #-1
    jnz -
    ; 16 instructions below
    ld DELAY_COUNTER_0
    addi #-1
    st DELAY_COUNTER_0
    jc rest_delay_1
    ld DELAY_COUNTER_1
    addi #-1
    st DELAY_COUNTER_1
    jc rest_delay_2
    ld DELAY_COUNTER_2
    addi #-1
    st DELAY_COUNTER_2
    jc rest_delay_3
    ld DELAY_COUNTER_3
    addi #-1
    st DELAY_COUNTER_3
    jc rest_delay

    jmp rest_delay_return

    ; make each loop be constant, regardless of how many nibbles were checked
rest_delay_1:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
rest_delay_2:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
rest_delay_3:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    jmp rest_delay


audio_play_tone:
    ; actual period in microseconds = (48 * (TONE_PERIOD + 1) + 60) / CPU_CLOCKS_PER_US
    ; TONE_PERIOD = (desired_period_us * CPU_CLOCKS_PER_US - 108) / 48
    lit #$4
    out #PORT_CONTROL

    ld TONE_PERIOD_0
    st TONE_COUNTER_0
    ld TONE_PERIOD_1
    st TONE_COUNTER_1
    ld TONE_PERIOD_2
    st TONE_COUNTER_2

    ; each pass through the loop takes 12 instructions
wait_low:
    ld TONE_COUNTER_0
    addi #-1
    st TONE_COUNTER_0
    jc period_delay_low_1
    ld TONE_COUNTER_1
    addi #-1
    st TONE_COUNTER_1
    jc period_delay_low_2
    ld TONE_COUNTER_2
    addi #-1
    st TONE_COUNTER_2
    jc wait_low
    jmp done_low

    ; make the period be constant, regardless of how many period nibbles were checked
period_delay_low_1:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
period_delay_low_2:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    jmp wait_low

done_low:
    lit #$C
    out #PORT_CONTROL
      
    ld TONE_PERIOD_0
    st TONE_COUNTER_0
    ld TONE_PERIOD_1
    st TONE_COUNTER_1
    ld TONE_PERIOD_2
    st TONE_COUNTER_2

    ; each pass through the loop takes 12 instructions
wait_high:
    ld TONE_COUNTER_0
    addi #-1
    st TONE_COUNTER_0
    jc period_delay_high_1
    ld TONE_COUNTER_1
    addi #-1
    st TONE_COUNTER_1
    jc period_delay_high_2
    ld TONE_COUNTER_2
    addi #-1
    st TONE_COUNTER_2
    jc wait_high
    jmp done_high

    ; make the period be constant, regardless of how many period nibbles were checked
period_delay_high_1:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
period_delay_high_2:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    jmp wait_high

done_high:         
    ; end of one tone period
    ld TONE_DURATION_0
    addi #-1
    st TONE_DURATION_0
    jc duration_delay_1
    ld TONE_DURATION_1
    addi #-1
    st TONE_DURATION_1
    jc duration_delay_2
    ld TONE_DURATION_2
    addi #-1
    st TONE_DURATION_2
    jc audio_play_tone
          
    jmp rest_delay

    ; make the period be constant, regardless of how many duration nibbles were checked
duration_delay_1:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
duration_delay_2:
    addi #0 ; NOP
    addi #0 ; NOP
    addi #0 ; NOP
    jmp audio_play_tone

; end music-stub.asm