#include "stdafx.h"
#include "Disassembly.h"
#include "MachineState.h"
#include "MainForm.h"

using namespace Simulator;
using namespace System::Windows::Forms;
using namespace System::IO;

Disassembly::Disassembly(MachineState^ machineState)
{	
	m_machineState = machineState;

	m_opcodeLookup[0x0] = InstructionData("JC", 2);
	m_opcodeLookup[0x1] = InstructionData("JNC", 2);
       m_opcodeLookup[0x2] = InstructionData("CMPI", 1);
	m_opcodeLookup[0x3] = InstructionData("CMPM", 2);
	m_opcodeLookup[0x4] = InstructionData("LIT", 1);
	m_opcodeLookup[0x5] = InstructionData("IN", 1);
	m_opcodeLookup[0x6] = InstructionData("LD", 2);
	m_opcodeLookup[0x7] = InstructionData("ST", 2);
	m_opcodeLookup[0x8] = InstructionData("JZ", 2);
	m_opcodeLookup[0x9] = InstructionData("JNZ", 2);
	m_opcodeLookup[0xA] = InstructionData("ADCI", 1);
	m_opcodeLookup[0xB] = InstructionData("ADCM", 2);
       m_opcodeLookup[0xC] = InstructionData("JMP", 2);
       m_opcodeLookup[0xD] = InstructionData("OUT", 1);
	m_opcodeLookup[0xE] = InstructionData("NORI", 1);
	m_opcodeLookup[0xF] = InstructionData("NORM", 2);	
}

int Disassembly::GetOpcodeSize(unsigned char opcode)
{
	if (m_opcodeLookup.ContainsKey(opcode))
	{
		return m_opcodeLookup[opcode].Length;
	}
	else
	{
		return 0;
	}
}

String^ Disassembly::GetOpcodeName(unsigned char opcode)
{
	if (m_opcodeLookup.ContainsKey(opcode))
	{
		return m_opcodeLookup[opcode].Mnemo;
	}
	else
	{
		return "???";
	}
}

String^ Disassembly::GetOpcodeDescription(unsigned char opcode)
{

	if (m_opcodeLookup.ContainsKey(opcode))
	{
		InstructionData data = m_opcodeLookup[opcode];
		String^ result = data.Mnemo;
		
		return result;
	}
	else
	{
		return "???";
	}
}


void Disassembly::FillDisassemblyListView(ListView^ listView, unsigned long startAddress, unsigned long pcAddress, unsigned long stkFrmAddr, Dictionary<unsigned long, bool>^ sourceBreakpoints, Dictionary<unsigned long, FileAndLine>^ addressToSource, Dictionary<String^, unsigned long>^ sourceStartAddress, bool ignoreSource)
{
	if (startAddress == 0xFFFFFFFF)
	{
		startAddress = 0;
	}

	const int linesToShow = 300;
	int currentInstructionIndex = -1;
	int instructionLength;
	//unsigned long nextAddress = startAddress;
       unsigned long nextAddress = pcAddress;
	unsigned long setBreakpointAddr = 0;

	listView->BeginUpdate();
	listView->Items->Clear();

	bool foundPC = false;

	// look for a source mapping
	bool foundSource = false;
	FileAndLine fileAndLine;
	if (!ignoreSource && addressToSource->TryGetValue(nextAddress, fileAndLine))
	{
		try
		{
			StreamReader^ sr = gcnew StreamReader(fileAndLine.filename);

			try
			{
				String^ line;
			 
				// Read and display lines from the file until the end of 
				// the file is reached.
				int lineNumber = 1;
				nextAddress = sourceStartAddress[fileAndLine.filename];
				while ( (line = sr->ReadLine()) != nullptr )
				{
					line = line->Replace("\t", "    ");
					ListViewItem^ lineItem = gcnew ListViewItem(line);
					lineItem->Tag = nextAddress; 

					listView->Items->Add(lineItem);

					if (lineNumber == fileAndLine.lineNumber && !foundPC)
                                   //if (pcAddress == nextAddress && !foundPC)
					{
						if (sourceBreakpoints->ContainsKey(nextAddress) && setBreakpointAddr != nextAddress)
						{
							lineItem->ImageKey = "breakpoint with arrow";
							setBreakpointAddr = nextAddress;
						}
						else
						{
							lineItem->ImageKey = "arrow";
						}
						currentInstructionIndex = lineItem->Index;
						foundPC = true;
					}
					else if (sourceBreakpoints->ContainsKey(nextAddress) && setBreakpointAddr != nextAddress)
					{
						lineItem->ImageKey = "breakpoint";
						setBreakpointAddr = nextAddress;
					}

					if (!sourceBreakpoints->ContainsKey(nextAddress))
					{	
						setBreakpointAddr = 0;
					}

					lineNumber++;

					int searchCount = 0;
					unsigned long testAddress = nextAddress;
					while (searchCount++ < 32) // skip up to 32 bytes
					{
						testAddress++;
						FileAndLine newFileAndLine;
						if (addressToSource->TryGetValue(testAddress, newFileAndLine))
						{
							if (newFileAndLine.lineNumber == lineNumber && 
								newFileAndLine.filename == fileAndLine.filename)
							{
								nextAddress = testAddress;
								break;
							}			
						}

					}

					if (searchCount > 32)
					{
						// didn't find it by looking up the key, so check all the values
						for each (KeyValuePair<unsigned long,FileAndLine> kvp in addressToSource)
						{
                                            FileAndLine fl = kvp.Value;
						  if (fl.lineNumber == lineNumber && 
						      fl.filename == fileAndLine.filename)
						  {
						      nextAddress = kvp.Key;
						      break;
						  }
						}
					}
				}

				foundSource = true;
			}
			finally
			{
				if ( sr )
				delete (IDisposable^)sr;	
			}
		}
		catch (Exception^)
		{
			MessageBox::Show( String::Format("Couldn't open source file {0}", fileAndLine.filename), "File Error" );
		}
	}
	
	if (!foundSource)
	{
		// disassembly
		for (int i=0; i<linesToShow; i++)
		{
			unsigned char b0 = m_machineState->PeekROM(nextAddress);
			unsigned char b1 = m_machineState->PeekROM(nextAddress+1);

			ListViewItem^ lineItem = gcnew ListViewItem(String::Format("{0:X3} ", nextAddress));
			lineItem->Tag = nextAddress;

			FileAndLine fileAndLine;
			if (addressToSource->TryGetValue(nextAddress, fileAndLine))
			{
				lineItem->Text += fileAndLine.sourceLabel;
				for (int i=fileAndLine.sourceLabel->Length; i<21; i++)
				{
					lineItem->Text += " ";
				}
			}
			else
			{
				lineItem->Text += "                     ";
			}

			unsigned char opcode = b0 >> 4;
			if (m_opcodeLookup.ContainsKey(opcode))
			{
				InstructionData% data = m_opcodeLookup[opcode];
				if (data.Length == 2)
				{
					unsigned long destination = ((m_machineState->PeekROM(nextAddress) & 0xF) << 8) | m_machineState->PeekROM(nextAddress+1);

					lineItem->Text += String::Format("{2:X2}{3:X2}    {0} ${1:X3}", data.Mnemo, destination, b0, b1);
				}
				else if (data.Length == 1)
				{
					lineItem->Text += String::Format("{2:X2}      {0} ${1:X1}\r\n", data.Mnemo, m_machineState->PeekROM(nextAddress) & 0xF, b0);
					
				}

				instructionLength = data.Length;
			}
			else
			{
				lineItem->Text += "???";
				instructionLength = 1;
			}

			listView->Items->Add(lineItem);

			if (nextAddress == pcAddress && !foundPC)
			{
				if (sourceBreakpoints->ContainsKey(nextAddress))
				{
					lineItem->ImageKey = "breakpoint with arrow";
				}
				else
				{
					lineItem->ImageKey = "arrow";
				}
				currentInstructionIndex = lineItem->Index;
				foundPC = true;
			}
			else if (sourceBreakpoints->ContainsKey(nextAddress))
			{
				lineItem->ImageKey = "breakpoint";
			}

			nextAddress += instructionLength;
		}
	}

	listView->EndUpdate();

	if (currentInstructionIndex >= 0)
	{
		listView->EnsureVisible(currentInstructionIndex);
	}

       //UpdateDisassemblyListView(listView, pcAddress, stkFrmAddr, sourceBreakpoints, addressToSource, true);
}

void Disassembly::UpdateDisassemblyListView(ListView^ listView, unsigned long pcAddress, unsigned long stackFrameAddress, Dictionary<unsigned long, bool>^ sourceBreakpoints, Dictionary<unsigned long, FileAndLine>^ addressToSource, bool scrollToCurrentInstruction)
{
	listView->BeginUpdate();

	bool foundPC = false;
	unsigned long setBreakpointAddr = 0;
	int currentInstructionIndex = -1;

	for each (ListViewItem^ item in listView->Items)
	{
		unsigned long address = safe_cast<unsigned long>(item->Tag);
		if (address == pcAddress && !foundPC)
		{
			if (sourceBreakpoints->ContainsKey(address) && setBreakpointAddr != address)
			{
				item->ImageKey = "breakpoint with arrow";
				setBreakpointAddr = address;
			}
			else
			{
				item->ImageKey = "arrow";
			}

			currentInstructionIndex = item->Index;

			foundPC = true;

		}
		else if (sourceBreakpoints->ContainsKey(address) && setBreakpointAddr != address)
		{
			item->ImageKey = "breakpoint";
			setBreakpointAddr = address;
		}
		else
		{
			item->ImageKey = nullptr;
		}

		if (!sourceBreakpoints->ContainsKey(address))
		{
			setBreakpointAddr = 0;
		}
	}

	if (scrollToCurrentInstruction)
	{
		if (currentInstructionIndex >= 0)
		{
			listView->EnsureVisible(currentInstructionIndex);
		}
	}

	listView->EndUpdate();
}
