#include "stdafx.h"
#include "LCD.h"

using namespace Simulator;
using namespace System::IO;
using namespace System::Globalization;
using namespace System::Windows::Forms;

LCD::LCD() :
	m_lcdPos(0),
       m_displayOn(false),
       m_ddramActive(true),
       m_4bitMode(false),
       m_4bitPhase(0),
       m_4bitNibble(0),
	m_lcdText("                                                                                ")
{
	
}

void LCD::WriteNibble(unsigned long port, unsigned char value)
{
    if (m_4bitMode)
    {
        if (m_4bitPhase == 0)
        {
            m_4bitNibble = value;
            m_4bitPhase++;
        }
        else
        {
            Write(port, (m_4bitNibble << 4) | value);
            m_4bitPhase = 0;
        }
    }
    else
    {
        // LCD is in 8-bit mode.
        Write(port, value << 4);
    }
}

// Store a byte 
void LCD::Write(unsigned long port, unsigned char value)
{
	if (port == 0)
	{
		// LCD control
		if (value == 0x01)
		{
			// clear
			m_lcdText = "                                                                                ";
			m_lcdPos = 0;
		}
		else if (value == 0x02)
		{
			// home
			m_lcdPos = 0;
		}
              else if ((value & 0xF8) == 0x08)
              {
                  m_displayOn = (value & 0x04) != 0;
              }
              else if ((value & 0xE0) == 0x20)
              {
                  m_4bitMode = (value & 0x10) == 0;
              }
              else if ((value & 0xC0) == 0x40)
              {
                  m_ddramActive = false;
              }
		else if (value & 0x80)
		{
	           // cursor pos
                  m_ddramActive = true;
		    m_lcdPos = value & 0x7F;
		}
	}
	else if (port == 1)
	{
            if (m_ddramActive)
            {
		// LCD data
		int stringPos = 0;
		if (m_lcdPos < 16)
		{
			stringPos = m_lcdPos;
		}
		else if (m_lcdPos >= 64 && m_lcdPos < 80)
		{
			stringPos = m_lcdPos - 48;
		}
		else
		{
			stringPos = -1;
		}

		String^ preText = gcnew String("");
		String^ postText = gcnew String("");
		if (stringPos > 0)
		{
			preText = m_lcdText->Substring(0, stringPos);
		}
		if (stringPos < 47)
		{
			postText = m_lcdText->Substring(stringPos + 1);
		}

              if (value < 8)
                  value += 48;

		m_lcdText = preText + Char::ConvertFromUtf32(value) + postText;	

		m_lcdPos++;

		/*if (m_lcdPos == 16)
		{
			m_lcdPos = 64;
		}
		else if (m_lcdPos == 88)
		{
			m_lcdPos = 0;
		}
              */
          }
	}
}