#pragma once

namespace Simulator {

	using namespace System;

	ref class LCD;

	public ref class MachineState 
	{
	public:
		MachineState(String^ programRomPath, String^ microRomPath, unsigned long bootAddress);

		void Reset();
		unsigned long GetMicroinstruction(unsigned long microAddress);
		unsigned long GetCurrentMicroaddress();
		unsigned long GetCurrentMicroinstruction();
		unsigned char GetCurrentInstruction();
              unsigned long GetCurrentMemoryAddress() { return ((Fetch & 0xF) << 8) | m_programROM[RegPC]; }
		unsigned char GetExecutingInstruction() { if (Phase == 1) return GetCurrentInstruction(); else return m_programROM[RegPC] >> 4; }
		unsigned char GetExecutingImmediate() { if (Phase == 1) return GetCurrentImmediate(); else return m_programROM[RegPC] & 0xF; }
		unsigned char GetCurrentImmediate();
		unsigned char PeekRAM(unsigned long address) { return m_RAM[address]; }
		unsigned char PeekROM(unsigned long address) { return m_programROM[address]; }
		//void Store(unsigned long address, unsigned char value);
              void SetButton(int index, bool pressed) { if (pressed) m_buttonState &= (~(1<<index)); else m_buttonState |= (1<<index); }

		String^ LCDText();

		unsigned long MicroStep();

		// 1-bit counter
		property unsigned char Phase;
		
		// 4-bit registers
		property unsigned char RegA;

		// 8-bit registers 
		property unsigned char Fetch;

		// 12-bit registers
		property unsigned long RegPC;

		// flags
		property bool FlagNotCarry;
		property bool FlagNotZero;

              // external peripherals
              property unsigned char Out0;
              property unsigned char Out1;

		enum class Dest
		{
			kLoadOut = 0,
			kLoadRegA
		};

		enum class BitPos
		{
			kLoadOut = 0,
			kOEImm,
			kOEIn,
			kOEALU,
			kWERAM,
			kCSRAM,
			kS0,
			kS1,
			kS2,
			kS3,
			kM,
			kNotCarryIn,
			kLoadFlags,
			kLoadA,
			kNotLoadPC,
			kIncPC
		};

		static bool IsBitZero(unsigned long bitfield, BitPos pos) { return (bitfield & (1 << (unsigned long)pos)) == 0; }
		static bool IsBitOne(unsigned long bitfield, BitPos pos) { return (bitfield & (1 << (unsigned long)pos)) != 0; }
		static unsigned long FieldValue(unsigned long bitfield, BitPos pos, int len);

	protected:
		~MachineState() {}
		bool LoadMicrorom(String^ romPath, int romIndex);	
		bool LoadProgramRom(String^ romPath);	

		static const int s_programRomSize = 8 * 1024;
		static const int s_ramSize = 4 * 1024;
		array<unsigned char>^ m_programROM;
		array<unsigned char>^ m_RAM;

              unsigned char m_buttonState;

		static const int s_microromSize = 2 * 1024;
		array<unsigned long>^ m_microrom;

		LCD^ m_lcd;

		unsigned long m_startAddress;

		unsigned char RunALU(unsigned long microOp, bool% carryOut, bool% overflowOut);
	};
}

