#include "stdafx.h"
#include "MicroDisassembly.h"
#include "MachineState.h"

using namespace Simulator;
using namespace System::Drawing;
using namespace System::Windows::Forms;

MicroDisassembly::MicroDisassembly(MachineState^ machineState)
{	
	m_machineState = machineState;
}

String^ MicroDisassembly::ConditionString(unsigned char bits, String^ name, bool isActiveLow)
{
	if (bits == 1 && !isActiveLow ||
		bits == 2 && isActiveLow)
	{
		return "!" + name + "&";
	} 
	else if (bits == 2 && !isActiveLow ||
			 bits == 1 && isActiveLow)
	{
		return name + "&";
	}
	else
	{
		return "";
	}

}

String^ MicroDisassembly::MicroOpDescription(unsigned long microOp, unsigned char phase)
{
	String^ result = "";
	String^ aluResult = "";
	String^ dataBus = "";

	unsigned char c = (unsigned char)MachineState::FieldValue(microOp, MachineState::BitPos::kNotCarryIn, 1);
	unsigned long function = MachineState::FieldValue(microOp, MachineState::BitPos::kS3, 4);
	
	if (phase == 0)
		result += "Fetch=ROM(PC)";

	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kCSRAM) && MachineState::IsBitOne(microOp, MachineState::BitPos::kWERAM))
	{
		dataBus = "RAM(imm<<8|ROM(PC))";
	}
	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kOEIn))
	{
		dataBus = "IN(imm)";
	}
	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kOEImm))
	{
		dataBus = "imm";
	}
	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kOEALU))
	{
		dataBus = "????";
	}

	if (MachineState::IsBitOne(microOp, MachineState::BitPos::kM))
	{
		// logical
		switch (function)
		{
			case 0:
				aluResult += "~A";
				break;
			case 1:
				aluResult += "~(A|" + dataBus + ")";
				break;
			case 3:
				aluResult += "0";
				break;
			case 5:
				aluResult += "~" + dataBus;
				break;
			case 6:
				aluResult += "A^" + dataBus;
				break;
			case 10:
				aluResult += dataBus;
				break;
			case 11:
				aluResult += "A&" + dataBus;
				break;
			case 12:
				aluResult += "$F";
				break;
			case 14:
				aluResult += "A|" + dataBus;
				break;
			case 15:
				aluResult += "A";
				break;
			default:
				aluResult += "ALU(A," + dataBus + "," + String::Format("${0:X1},{1},1", function, c) + ")";
				break;
		}
	}
	else
	{
		// arithmetic
		switch (function)
		{
			case 0:
				aluResult += c ? "A" : "A+1";
				break;
			case 3:
				aluResult += c ? "-1" : "0";
				break;
			case 6:
				aluResult += c ? "A-" + dataBus + "-1": "A-" + dataBus;
				break;
			case 9:
				aluResult += c ? "A+" + dataBus: "A+" + dataBus+ "+1";
				break;
			case 12:
				aluResult += c ? "A+A" : "A+A+1" ;
				break;
			case 15:
				aluResult += c ? "A-1" : "A";
				break;
			default:
				aluResult += "ALU(A," + dataBus + "," + String::Format("${0:X1},{1},0", function, c) + ")";
				break;
		}
	}

	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kOEALU))
	{
		dataBus = aluResult;
	}

	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kLoadA))
	{
		result += "A=" + aluResult;
	}

	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kLoadOut))
	{
		result += "OUT(imm)=" + dataBus;
	}

	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kCSRAM) && MachineState::IsBitZero(microOp, MachineState::BitPos::kWERAM))
	{
		result += "RAM(imm<<8|ROM(PC))=" + dataBus;
	}

	// flags
	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kLoadFlags))
	{
		if (result->Length == 0)
		{
			result = aluResult;
		}

		result += "; loadFlags";
	}

	// PC
	if (MachineState::IsBitZero(microOp, MachineState::BitPos::kNotLoadPC))
		result += "; PC=imm<<8|ROM(PC)";
	else if (MachineState::IsBitOne(microOp, MachineState::BitPos::kIncPC))
		result += "; PC++";

	return result;
}

String^ MicroDisassembly::RightALUString(unsigned long microOp)
{
	return "Right";
}

void MicroDisassembly::FillDisassemblyListView(ListView^ listView, Dictionary<unsigned long, List<unsigned long>^>^ microSourceBreakpoints)
{
	listView->BeginUpdate();
	listView->Items->Clear();

	unsigned char currentPhase = m_machineState->Phase;
	unsigned long currentMicroAddress = m_machineState->GetCurrentMicroaddress();

	for (unsigned char phase=0; phase<2; phase++)
	{
		m_microInstructionLookup.Clear();
			
		for (unsigned char c=0; c<2; c++)
		{
			for (unsigned char z=0; z<2; z++)
			{				
				unsigned long microOp = m_machineState->GetMicroinstruction(m_machineState->GetExecutingInstruction() * 8 + c * 4 + z * 2 + phase);
							
				if (!m_microInstructionLookup.ContainsKey(microOp))
				{
					m_microInstructionLookup[microOp] = MicroInstructionData();
				}

				MicroInstructionData data = m_microInstructionLookup[microOp];
				data.Carry |= 1<<c;
				data.Zero |= 1<<z;
				m_microInstructionLookup[microOp] = data;
			}
		}

		bool lastPhase = true;

		for each (KeyValuePair<unsigned long,MicroInstructionData> entry in m_microInstructionLookup)
		{
			ListViewItem^ lineItem = gcnew ListViewItem();
			unsigned long microProgramBaseAddress = (m_machineState->GetExecutingInstruction() << 3) | phase;
			unsigned long microProgramConditionCodes = (entry.Value.Carry << 2) | (entry.Value.Zero << 0);
		
			lineItem->Tag = gcnew KeyValuePair<unsigned long, unsigned long>(microProgramBaseAddress, microProgramConditionCodes);
			lineItem->BackColor = phase % 2 ? Color::FromArgb(240, 240, 240) : Color::White;

			String^ conditionString = "";

			conditionString += ConditionString(entry.Value.Carry, "C", true);
			conditionString += ConditionString(entry.Value.Zero, "Z", false);

			// strip trailing ampersand
			if (conditionString->EndsWith("&"))
			{
				conditionString = "if (" + 
					conditionString->Substring(0, conditionString->Length-1) + ") ";

				if (conditionString->Length % 2 == 1)
				{
					conditionString += " ";
				}
			}

			String^ operationString = MicroOpDescription(entry.Key, phase);
			
			// strip out do-nothing data move if there's anything else being done
			if (operationString->StartsWith("A <- A; "))
			{
				operationString = operationString->Substring(8);
			}

			lineItem->Text = String::Format("{0:X1} {1}{2}", phase, conditionString, operationString);

			if (phase == currentPhase &&
				(entry.Value.Carry & (1 << ((currentMicroAddress >> 2) & 0x1))) &&
				(entry.Value.Zero & (1 << ((currentMicroAddress >> 1) & 0x1))))
			{
				lineItem->ImageKey = "arrow";

				if (microSourceBreakpoints->ContainsKey(microProgramBaseAddress))
				{
					List<unsigned long>^ conditionCodes = microSourceBreakpoints[microProgramBaseAddress];
					for each (unsigned long conditionCode in conditionCodes)
					{
						if (conditionCode == microProgramConditionCodes)
						{
							lineItem->ImageKey = "breakpoint with arrow";
							break;
						}
					}
				}
			}
			else if (microSourceBreakpoints->ContainsKey(microProgramBaseAddress))
			{
				List<unsigned long>^ conditionCodes = microSourceBreakpoints[microProgramBaseAddress];
				for each (unsigned long conditionCode in conditionCodes)
				{
					if (conditionCode == microProgramConditionCodes)
					{
						lineItem->ImageKey = "breakpoint";
						break;
					}
				}
			}


			lastPhase = false;

			listView->Items->Add(lineItem);
		}

		// if all CC's do a nextOp on this phase, then don't list the remaining phases
		if (lastPhase)
		{
			break;
		}
	}

	listView->EndUpdate();
}